from __future__ import annotations

from django.conf import settings
from django.core.management import BaseCommand

from schematools.contrib.django.models import Scope
from schematools.loaders import get_schema_loader
from schematools.types import Scope as ScopeSchema


class Command(BaseCommand):
    help = "Import all known scopes from Amsterdam schema files."
    requires_system_checks = []

    def add_arguments(self, parser) -> None:
        parser.add_argument("scope", nargs="*", help="Local scope files to import")
        parser.add_argument(
            "--schema-url",
            default=settings.SCHEMA_URL,
            help=f"Schema URL (default: {settings.SCHEMA_URL})",
        )

    def handle(self, *args, **options) -> None:
        if options["scope"]:
            scopes = self.import_from_files(options["scope"])
        else:
            scopes = self.import_from_url(options["schema_url"])

        if scopes:
            self.stdout.write(f"Imported scopes: {len(scopes)}")
        else:
            self.stdout.write("No new scopes imported.")

    def import_from_files(self, scope_files: list[str]) -> list[Scope]:
        scopes = []
        for filename in scope_files:
            schema = ScopeSchema.from_file(filename)
            scope = self._import(schema)
            if scope is not None:
                scopes.append(scope)

        return scopes

    def import_from_url(self, schema_url: str) -> list[Scope]:
        """Import all schema definitions from an URL"""
        self.stdout.write(f"Loading scopes from {schema_url}")
        loader = get_schema_loader(schema_url)
        scopes = []

        for _, schema in loader.get_all_scopes().items():
            scope = self._import(schema)
            if scope is not None:
                scopes.append(scope)
        return scopes

    def _import(self, schema: ScopeSchema) -> Scope | None:
        self.stdout.write(f"* Processing {schema.name}")
        id = schema.id
        try:
            scope = Scope.objects.get(id=id)
        except Scope.DoesNotExist:
            scope = Scope.create_for_schema(schema)
            self.stdout.write(f"   Created {id}")
            return scope
        else:
            updated = scope.save_for_schema(schema)
            if updated:
                self.stdout.write(f"    Updated {schema.name}")
                return scope

        return None



# Generated by Django 3.2.19 on 2023-07-09 13:22
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0015_alter_dataset_id_alter_datasetfield_id_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="view_data",
            field=models.TextField(blank=True, null=True, verbose_name="View SQL"),
        ),
    ]



from __future__ import annotations


class ParserError(ValueError):
    """Exception to indicate the parsing failed."""


class SchemaObjectNotFound(ValueError):
    """Field does not exist."""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class ViewObjectNotFound(ValueError):
    """SQL view does not exist."""


class DatasetNotFound(SchemaObjectNotFound):
    """The dataset could not be found."""


class DatasetVersionNotFound(SchemaObjectNotFound):
    """The version could not be found."""


class DatasetTableNotFound(SchemaObjectNotFound):
    """The table could not be found."""


class DatasetFieldNotFound(SchemaObjectNotFound):
    """The field could not be found."""


class IncompatibleMetaschema(Exception):
    """This package version of schema-tools
    is being used with a metaschema that it is
    not compatible with."""


class IncompatibleDataset(Exception):
    """Datasets to compare should be the same. Cannot compare different datasets."""


class PendingMetaschemaDeprecation(PendingDeprecationWarning):
    """The used metaschema is marked for deprecation."""


class DuplicateScopeId(ValueError):
    """The ID is already used in another scope."""


class DuplicateProfileId(ValueError):
    """The ID is already used for another profile."""


class ScopeNotFound(SchemaObjectNotFound):
    """The scope could not be found."""


class LoaderNotFound(Exception):
    """The loader could not be found."""



from __future__ import annotations

from factory import Faker
from faker.providers import BaseProvider


class NullerProvider(BaseProvider):  # noqa: D101
    """Provider for an null value."""

    def nuller(self) -> None:  # noqa: D102
        return None


Faker.add_provider(NullerProvider, locale="nl_NL")



from __future__ import annotations

import re

from django.apps import apps
from django.core.management import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.migrations import Migration
from django.db.migrations.autodetector import MigrationAutodetector
from django.db.migrations.graph import MigrationGraph
from django.db.migrations.questioner import InteractiveMigrationQuestioner
from django.db.migrations.state import ModelState, ProjectState

from schematools.contrib.django.factories import DjangoModelFactory
from schematools.contrib.django.models import Dataset
from schematools.types import DatasetTableSchema


def migrate(
    command: BaseCommand,
    current_dataset: Dataset,
    updated_dataset: Dataset,
    current_table: DatasetTableSchema,
    updated_table: DatasetTableSchema,
    real_apps: list[str],
    dry_run: bool = True,
    database: str = DEFAULT_DB_ALIAS,
    project_state: ProjectState = None,
):
    """
    Creates a project state for both the current dataset and the updated dataset.
    """
    # Generate a full project state that incorporates the table versions
    # The used Django migration-API calls were inspired by reading
    # the 'manage.py makemigrations' and 'manage.py sqlmigrate' command changes.
    command.stdout.write(f"* Processing table {updated_table.id}")
    base_state = get_base_project_state(command, current_dataset, current_table.id, real_apps)
    current_state = project_state or get_versioned_project_state(
        base_state, current_dataset, current_table
    )
    updated_state = get_versioned_project_state(base_state, updated_dataset, updated_table)

    # Clear any models from the app cache to avoid confusion
    for vmajor in current_dataset.schema.versions:
        app_label = f"{current_dataset.schema.id}_{vmajor}"
        if app_label in apps.all_models:
            del apps.all_models[app_label]
        if app_label in apps.app_configs:
            del apps.app_configs[app_label]
    apps.clear_cache()

    start_state = current_state

    # Let the migration engine perform its magic, similar to `manage.py makemigrations`:
    migrations = get_migrations(
        current_state,
        updated_state,
        app_name=f"{current_dataset.schema.id}_{current_dataset.schema.default_version}",
    )
    if not migrations:
        command.stdout.write(f"  No changes detected for table {current_table.id}")
        return start_state

    # Generate SQL per migration file, just like `manage.py sqlmigrate` does:
    connection = connections[database]
    for _app, app_migrations in migrations.items():
        for migration in app_migrations:
            start_state = execute_migration(
                command, connection, start_state, migration, dry_run=dry_run
            )
    return start_state


def drop_table(table_id):
    """Drop a database table."""
    connection = connections[DEFAULT_DB_ALIAS]
    with connection.schema_editor() as schema_editor:
        schema_editor.execute(f"DROP TABLE IF EXISTS {table_id};")


def get_base_project_state(
    command: BaseCommand, dataset_model: Dataset, exclude_table: str, real_apps: list[str]
) -> ProjectState:
    """Generate the common/shared project state.

    This includes all other models of the dataset, as those may be referenced by relations.
    It excludes the versioned table, since that will differ.
    """
    if command.verbosity >= 2:
        command.stdout.write(f"-- Building shared state for {dataset_model.name}")

    project_state = ProjectState(real_apps=set(real_apps))

    # Generate model states for all other tables in the dataset
    for table in dataset_model.schema.get_tables(include_nested=True, include_through=True):
        # Exclude the actual table that changes, including any nested/through tables.
        if table.id == exclude_table or (
            table.has_parent_table and table.parent_table.id == exclude_table
        ):
            continue

        project_state.add_model(get_model_state(dataset_model, table))

    return project_state


def get_versioned_project_state(
    base_state: ProjectState, dataset_model: Dataset, table: DatasetTableSchema
) -> ProjectState:
    """Generate the final project state.
    This clones the base state, so other related models are only created once.
    """
    project_state = base_state.clone()
    project_state.add_model(get_model_state(dataset_model, table))

    # Add any nested tables and through tables,
    # that are also part of this table schema.
    for field in table.fields:
        if (through_table := field.through_table) is not None:
            project_state.add_model(get_model_state(dataset_model, through_table))
        elif (nested_table := field.nested_table) is not None:
            project_state.add_model(get_model_state(dataset_model, nested_table))

    return project_state


def get_model_state(dataset_model: Dataset, table: DatasetTableSchema, managed=True) -> ModelState:
    """Generate the model state for a table."""
    # The migration-engine will only consider models that have "managed=True".
    # This is turned off by default for the DjangoModelFactory.build_model() logic,
    # and needs to be overwritten here (patching model._meta and its original_attrs is nasty).
    factory = DjangoModelFactory(dataset_model)
    model_class = factory.build_model(table, meta_options={"managed": managed})

    # Generate the model. exclude_rels=True because M2M-through tables are generated manually.
    return PatchedModelState.from_model(model_class, exclude_rels=True)


def get_migrations(
    state1: ProjectState, state2: ProjectState, app_name: str
) -> dict[str, list[Migration]]:
    """Generate a migration object for the given table versions."""
    detector = MigrationAutodetector(
        from_state=state1,
        to_state=state2,
        questioner=InteractiveMigrationQuestioner(specified_apps=[app_name]),
    )
    # The dependency graph remains empty here, as we assume all other models are still
    # in place. We only compare the changes between 2 models of the same table.
    dependency_graph = MigrationGraph()
    return detector.changes(
        dependency_graph,
        trim_to_apps=[app_name],
        migration_name=app_name,
    )


def execute_migration(
    command: BaseCommand,
    connection: BaseDatabaseWrapper,
    start_state: ProjectState,
    migration: Migration,
    dry_run: bool = True,
) -> ProjectState:
    """Print the SQL statements for a migration"""
    with connection.schema_editor(collect_sql=True, atomic=migration.atomic) as schema_editor:
        command.stdout.write(f"-- Migration {migration.name} for dataset {migration.app_label}")
        try:
            start_state = migration.apply(start_state, schema_editor, collect_sql=True)
        except Exception:
            # On crashes, still show the generated statements so far
            command.stdout.write(
                "\n".join(_filter_alter_type_statements(schema_editor.collected_sql))
            )
            raise

        # Filter out ALTER COLUMN statements
        collected_sql = _filter_alter_type_statements(schema_editor.collected_sql)
        # Escape % signs
        collected_sql = _escape_comment_statements(collected_sql)

        # If we've only got comments left, skip this migration
        if all(statement.startswith("--") for statement in collected_sql):
            command.stdout.write("-- No actual SQL statements generated, skipping this migration")
            return start_state

        if dry_run:
            command.stdout.write("-- DRY RUN - the following would be executed:")
            command.stdout.write("\n".join(collected_sql))
        else:
            schema_editor.collect_sql = False
            schema_editor.execute("\n".join(collected_sql))
    return start_state


def _filter_alter_type_statements(sql: list) -> list:
    """Filter ALTER TYPE statements from the generated SQL since this causes issues
    with our views. This function can be removed once we either:
        - Drop our views before running the ALTER TYPE statements and recreate them
        - Patch the migration logic in Django to not create an ALTER TYPE statement when
          adding or changing a db_comment
    """
    return [s for s in sql if not re.search(r"ALTER COLUMN.+TYPE", s)]


def _escape_comment_statements(sql: list) -> list:
    """Escape percent signs in comments from the generated SQL since this causes issues
    when executing this on the database.
    """
    return [s.replace("%", "%%") if re.search(r"COMMENT.+", s) else s for s in sql]


class PatchedModelState(ModelState):
    """A workaround to avoid breaking migration rendering."""

    def clone(self):
        """Return an exact copy of this ModelState."""
        # Workaround for Django issue. The fields get bound to a model during the first
        # migration operation, which breaks reusing them in migrations.
        # Quick fix is to deep-clone the fields too:
        self.fields = {name: field.clone() for name, field in self.fields.items()}
        return super().clone()



"""Create GRANT statements to give roles very specific access to the database."""

from __future__ import annotations

import logging

from pg_grant import PgObjectType, parse_acl_item, query
from pg_grant.sql import _Grant, grant, revoke
from sqlalchemy import Connection, event, text
from sqlalchemy.engine import Engine

from schematools.permissions import PUBLIC_SCOPE
from schematools.types import (
    DatasetFieldSchema,
    DatasetSchema,
    DatasetTableSchema,
    PermissionLevel,
    ProfileSchema,
    Scope,
)

# Create a module-level logger, so calling code can
# configure the logger, if needed.
logger = logging.getLogger(__name__)

existing_roles = set()  # note: used as global cache!
existing_sequences = {}

PUBLIC_SCOPE_OBJECT = Scope({"id": PUBLIC_SCOPE})
PUBLIC_SCOPES = {PUBLIC_SCOPE_OBJECT, PUBLIC_SCOPE}


def introspect_permissions(engine: Engine, role: str) -> None:
    """Shows the table permissions."""
    with engine.connect() as conn:
        schema_relation_infolist = query.get_all_table_acls(conn, schema="public")

    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    logger.info(
                        'role "%s" has privileges %s on table "%s"',
                        role,
                        ",".join(acl.privs),
                        schema_relation_info.name,
                    )


def revoke_permissions(engine: Engine, role: str, verbose: int = 0) -> None:
    """Revoke all privileges for the indicated role."""
    grantee = role
    with engine.connect() as conn:
        schema_relation_infolist = query.get_all_table_acls(conn, schema="public")

    revoke_statements = []
    for schema_relation_info in schema_relation_infolist:
        if schema_relation_info.acl:
            acl_list = [parse_acl_item(item) for item in schema_relation_info.acl]
            for acl in acl_list:
                if acl.grantee == role:
                    revoke_statements.append(
                        revoke("ALL", PgObjectType.TABLE, schema_relation_info.name, grantee)
                    )

    with engine.begin() as conn:
        for revoke_statement in revoke_statements:
            if verbose:
                logger.info(
                    'revoking ALL privileges of role "%s" on table "%s"',
                    role,
                    revoke_statement.target,
                )

            conn.execute(revoke_statement)


def apply_schema_and_profile_permissions(
    engine: Engine,
    schemas: DatasetSchema | dict[str, DatasetSchema],
    profiles: list[ProfileSchema] | None,
    *,
    only_role: str | None = None,
    only_scope: str | None = None,
    set_read_permissions: bool = True,
    set_write_permissions: bool = True,
    dry_run: bool = False,
    create_roles: bool = False,
    revoke: bool = False,
    verbose: int = 0,
    additional_grants: tuple[str] = (),
    all_scopes: list[Scope] | None = None,
) -> None:
    """Apply permissions for schema and profile.

    Read permissions are granted to roles 'scope_X', where X are scopes found in Amsterdam Schema.
    Write permissions are granted to roles 'write_Y', where Y are dataset ids,
    for all tables belonging to the dataset.
    Revoke old privileges before assigning new in case new privileges are more restrictive.
    """
    datasets = {schemas.id: schemas} if isinstance(schemas, DatasetSchema) else schemas

    if verbose:
        event.listen(engine, "before_cursor_execute", _before_cursor_execute)

    with engine.connect() as conn:
        try:
            if datasets:
                if revoke:
                    # Remove all privileges first
                    revoke_dataset = schemas if isinstance(schemas, DatasetSchema) else None
                    revoke_schema_permissions(
                        conn, revoke_dataset, only_role, dry_run, verbose=verbose
                    )

                if create_roles:
                    for scope in all_scopes or []:
                        role = _scope_to_role(scope)
                        filtered_role = f"{role}.filtered"
                        _create_role_if_not_exists(conn, role, verbose=verbose, dry_run=dry_run)
                        _create_role_if_not_exists(
                            conn, filtered_role, inherits=role, verbose=verbose, dry_run=dry_run
                        )

                # Apply privileges for all datasets, or the selected dataset.
                apply_schema_permissions(
                    conn,
                    datasets,
                    only_role,
                    only_scope,
                    set_read_permissions,
                    set_write_permissions,
                    dry_run=dry_run,
                    create_roles=create_roles,
                    verbose=verbose,
                )

                if profiles:
                    apply_profile_permissions(
                        conn,
                        profiles,
                        datasets,
                        only_role,
                        only_scope,
                        dry_run=dry_run,
                        create_roles=create_roles,
                        verbose=verbose,
                    )

            if additional_grants:
                apply_additional_grants(
                    conn, additional_grants, dry_run=dry_run, create_roles=False, verbose=verbose
                )

            conn.commit()
        except Exception:
            conn.rollback()
            logger.warning("Session rolled back")
            raise
        finally:
            if verbose:
                event.remove(engine, "before_cursor_execute", _before_cursor_execute)


def _before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
    """Report notices raised by the 'RAISE NOTICE' statements."""
    raw_connection = cursor.connection
    if not raw_connection._notice_handlers:
        raw_connection.add_notice_handler(_on_cursor_notice)


def _on_cursor_notice(diagnostic):
    logger.info(diagnostic.message_primary)


def _collect_dataset_write_grants(conn: Connection, ams_schema: DatasetSchema) -> list[_Grant]:
    """Sets write permissions for the indicated dataset."""
    grantee = f"write_{ams_schema.db_name}"
    all_grants = []
    for table in ams_schema.get_all_tables(include_nested=True, include_through=True):
        all_grants.extend(
            _build_table_grants(
                conn,
                table,
                privileges=["SELECT", "INSERT", "UPDATE", "DELETE", "TRUNCATE", "REFERENCES"],
                grantees=[grantee],
            )
        )
    return all_grants


def _collect_dataset_grants(
    conn: Connection,
    dataset: DatasetSchema,
    only_role: str | None = None,
    only_scope: str | None = None,
) -> list[_Grant]:
    """Returns all scopes that should be applied to the tables of a dataset.

    Args:
        session: the SQL Alchemy session; used for getting the sequence names
        dataset: the amsterdam schema that needs to be processed
        only_role: only return grants for a specific role, none means all roles.
        only_scope: only return grants for a specific scope, none means all scopes.

        The grants will be applied according to the configuration of the auth scopes
        in the amsterdam schema.
        If not auth scopes are defined, all tables get the `scope_openbaar` grant.
        If only the dataset has as scope `foo`, alle tables get the `scope_foo` grant.
        If a table has as scope `bar`, this overrules the dataset scope,
        so this table gets the `scope_bar` grant.
        If one or more fields have a scope, this scope overrules both the
        dataset and the table scope. The other fields in the same table
        get a `scope_openbaar` grant in that case.
        If a 1-N relation field has a scope, the foreign key field get the associated
        grant. In case this relation field is of type `object`` (e.g. for temporal fields),
        the additional columns (usually identificatie/volgnummer postfixed) get the same grant.
        If NM and nested relation fields (type `array` in the schema) have a scope `bar`
        the associated sub-table gets the grant `scope_bar`.

    Returns:
        all_scopes (list): Contains a list of scopes with privileges and grants:

        [
            grant(["SELECT"],           PgObjectType.TABLE, "table1", "scope_openbaar"),
            grant(["SELECT (columnA)"], PgObjectType.TABLE, "table2", "scope_openbaar"),
            grant(["SELECT (columnB)"], PgObjectType.TABLE, "table2", "scope_A"),
            grant(["SELECT (columnB)"], PgObjectType.TABLE, "table2", "scope_B"]),
        ]
    """
    grants = []
    for table in dataset.get_all_tables(include_nested=True, include_through=True):
        table_scopes = (table.scopes - PUBLIC_SCOPES) or dataset.scopes
        fields = [
            field
            for field in table.get_fields(include_subfields=True)
            if not field.type.endswith("#/definitions/schema")
        ]

        # First process all fields, to know if any fields has a non-public scope
        column_scopes = _get_column_level_scopes(fields)
        if column_scopes:
            # When some fields have a specific scope,
            # grant statements need to be generated for each individual field.
            # After all, "GRANT SELECT ON 'table'" would bypass field-level access.
            for field in fields:
                if field.nm_relation or field.is_nested_table:
                    # field is not in view when nm or nested
                    continue

                # Column-level read permissions
                scopes = column_scopes.get(field.db_name, table_scopes)
                if grantees := _filter_grantees(scopes, only_role, only_scope):
                    grants.extend(_build_field_grants(conn, field, grantees))
        elif grantees := _filter_grantees(table_scopes, only_role, only_scope):
            # Table-level read permissions
            grants.extend(_build_table_grants(conn, table, ["SELECT"], grantees))

    return grants


def _collect_profile_grants(
    conn: Connection,
    profile: ProfileSchema,
    datasets: dict[str, DatasetSchema],
    only_role: str | None = None,
    only_scope: str | None = None,
) -> list[_Grant]:
    """Tell which grants a profile should given.

    Args:
        conn: SQLAlchemy connection.
        profile: The profile to calculate GRANT statements for
        datasets: Associated datasets touched by the profile.
        only_role: The role to filter grants by.
        only_scope: The scope to filter grants by.
    """
    grants = []
    grantees = _filter_grantees(profile.scopes, only_role=only_role, only_scope=only_scope)
    if not grantees:
        return []

    for profile_dataset in profile.datasets.values():
        dataset = datasets[profile_dataset.id]

        if profile_dataset.permissions.level >= PermissionLevel.LETTERS:
            # READ access on dataset level.
            # Giving read access to a whole dataset.
            for table in dataset.get_all_tables(include_nested=True, include_through=True):
                grants.extend(_build_table_grants(conn, table, ["SELECT"], grantees))
        else:
            dataset_tables = {table.id: table for table in dataset.tables}
            for profile_table in profile_dataset.tables.values():
                table = dataset_tables[profile_table.id]
                table_grantees = (
                    [f"{s}.filtered" for s in grantees]
                    if profile_table.mandatory_filtersets
                    else grantees
                )

                if profile_table.permissions:
                    # READ access on table level.
                    grants.extend(_build_table_grants(conn, table, ["SELECT"], table_grantees))
                elif profile_table.fields:
                    # READ access on field level.
                    fields_by_name = {
                        (f"{f.parent_field.id}.{f.id}" if f.parent_field is not None else f.id): f
                        for f in table.get_fields(include_subfields=True)
                    }

                    for field_name, field_permissions in profile_table.fields.items():
                        if field_permissions:  # sanity check
                            grants.extend(
                                _build_field_grants(
                                    conn, fields_by_name[field_name], table_grantees
                                )
                            )
    return grants


def _get_column_level_scopes(fields: list[DatasetFieldSchema]) -> dict[str, frozenset[Scope]]:
    """Tell whether there are fields that have an explicit scope."""
    column_scopes = {}
    for field in fields:
        # Object type relations have subfields, in that case
        # the auth scope on the relation is leading.
        field_scopes = field.scopes - PUBLIC_SCOPES

        if field.is_subfield:
            field_scopes = (field.parent_field.scopes - PUBLIC_SCOPES) or field_scopes

        if field_scopes:
            column_scopes[field.db_name] = field_scopes

    return column_scopes


def _build_table_grants(
    conn: Connection,
    table: DatasetTableSchema,
    privileges: list[str],
    grantees: list[str],
) -> list[_Grant]:
    """Build the SELECT grants for accessing a full table."""
    grants = [
        grant(
            privileges,
            type=PgObjectType.TABLE,
            target=table.db_name,
            grantee=grantee,
            schema="public",
        )
        for grantee in grantees
    ]

    if sequence_name := _get_sequence_name(conn, table):
        grants.extend(
            grant(
                ["USAGE" if "INSERT" in privileges else "SELECT"],
                type=PgObjectType.SEQUENCE,
                target=sequence_name,
                grantee=grantee,
                schema="public",
            )
            for grantee in grantees
        )
    return grants


def _build_field_grants(
    conn: Connection, field: DatasetFieldSchema, grantees: list[str]
) -> list[_Grant]:
    """Build the SELECT grants for accessing a field."""
    table = field.table
    grants = [
        grant(
            # NB. space after SELECT is significant!
            privileges=[f"SELECT ({field.db_name})"],
            type=PgObjectType.TABLE,
            target=table.db_name,
            grantee=grantee,
            schema="public",
        )
        for grantee in grantees
    ]

    # Get PostgreSQL generated sequence for 'id' column that Django added.
    if field.is_primary and (sequence_name := _get_sequence_name(conn, table)):
        grants.extend(
            grant(
                privileges=["SELECT"],
                type=PgObjectType.SEQUENCE,
                target=sequence_name,
                grantee=grantee,
                schema="public",
            )
            for grantee in grantees
        )
    return grants


def _collect_additional_grants(additional_grants: tuple[str]) -> list[_Grant]:
    """Parse the additional grants syntax into GRANT statements."""
    all_grants = []
    for additional_grant in additional_grants:
        try:
            table_name, grant_params = additional_grant.split(":")
            privileges_str, grantees_str = grant_params.split(";")
            privileges = privileges_str.split(",")
            grantees = grantees_str.split(",")
        except ValueError:
            logger.error(
                "Incorrect grant definition: `%r`,"
                "grant should have format"
                "<table_name>:<privilege_1>[,<privilege_n>]*;<grantee_1>[,grantee_n]*",
                additional_grant,
            )
            raise  #  re-raise to trigger rollback

        all_grants.extend(
            grant(
                privileges,
                PgObjectType.TABLE,
                table_name,
                _grantee,
                schema="public",
            )
            for _grantee in grantees
        )

    return all_grants


def apply_schema_permissions(
    conn: Connection,
    datasets: dict[str, DatasetSchema],
    only_role: str | None,
    only_scope: str | None,
    set_read_permissions: bool,
    set_write_permissions: bool,
    dry_run: bool,
    create_roles: bool = False,
    verbose: int = 0,
) -> None:
    """Create and set the ACL for automatically generated roles based on Amsterdam Schema.

    Read permissions are granted to roles 'scope_X', where X are scopes found in Amsterdam Schema
    Write permissions are granted to roles 'write_Y', where Y are dataset ids,
    for all tables belonging to the dataset.
    Revoke old privileges before assigning new in case new privileges are more restrictive.
    """
    for dataset in datasets.values():
        all_grants = (
            _collect_dataset_grants(conn, dataset, only_role=only_role, only_scope=only_scope)
            if set_read_permissions
            else []
        )
        if set_write_permissions:
            all_grants.extend(_collect_dataset_write_grants(conn, dataset))

        _execute_grants(
            conn,
            all_grants,
            dry_run=dry_run,
            create_roles=create_roles,
            verbose=verbose,
        )


def apply_profile_permissions(
    conn: Connection,
    profiles: list[ProfileSchema],
    datasets: dict[str, DatasetSchema],
    only_role: str | None = None,
    only_scope: str | None = None,
    dry_run: bool = False,
    create_roles: bool = False,
    verbose: int = 0,
) -> None:
    """Create an ACL from profile list."""
    dataset_ids = set(datasets.keys())
    profiles = [
        p
        for p in profiles
        if dataset_ids.intersection(p.datasets.keys())
        and only_scope is None
        or only_scope in p.scopes
    ]
    if not profiles:
        return

    # Retrieve all grants for the profiles
    all_grants = []
    for profile in profiles:
        all_grants.extend(
            _collect_profile_grants(
                conn, profile, datasets, only_role=only_role, only_scope=only_scope
            )
        )

    _execute_grants(conn, all_grants, dry_run=dry_run, create_roles=create_roles, verbose=verbose)


def apply_additional_grants(
    conn: Connection,
    additional_grants: tuple[str],
    dry_run: bool = False,
    create_roles: bool = False,
    verbose: int = 0,
):
    """
    Parse manaully defined grants.

    Args:
        additional_grants: tuple with the following structure:
            <table_name>:<privilege_1>[,<privilege_n>]*;<grantee_1>[,grantee_n]*
    """
    all_grants = _collect_additional_grants(additional_grants)
    _execute_grants(conn, all_grants, dry_run=dry_run, create_roles=create_roles, verbose=verbose)


def revoke_schema_permissions(
    conn: Connection,
    only_dataset: DatasetSchema | None = None,
    only_role: str | None = None,
    dry_run: bool = False,
    verbose: int = 1,
):
    """Revoke all privileges that may have been previously granted.

    This is about grants to the scope_* and write_* roles.
    If dataset is provided, revoke only rights to the tables belonging to
    dataset.
    """
    pg_schema = "public"
    db_role_names = _get_all_role_names(conn) if only_role is None else [only_role]

    revoke_statements = []
    if only_dataset:
        # for a single dataset
        for table in only_dataset.tables:
            revoke_statements.extend(
                f'REVOKE ALL PRIVILEGES ON "{pg_schema}.{table.db_name}" FROM "{role_name}"'
                for role_name in db_role_names
            )
            if sequence_name := _get_sequence_name(conn, table):
                revoke_statements.extend(
                    f'REVOKE ALL PRIVILEGES ON SEQUENCE "{pg_schema}.{sequence_name}"'
                    f' FROM "{role_name}"'
                    for role_name in db_role_names
                )
    else:
        # For all datasets
        revoke_statements = [
            f'REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA {pg_schema} FROM "{role_name}"'
            for role_name in db_role_names
        ] + [
            f'REVOKE ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA {pg_schema} FROM "{role_name}"'
            for role_name in db_role_names
        ]

    # Execute all for this role
    for statement in revoke_statements:
        _execute_grant(conn, statement, verbose=verbose, dry_run=dry_run)


def _execute_grants(
    conn,
    all_grants: list[_Grant],
    dry_run: bool,
    create_roles: bool = False,
    verbose: int = 0,
) -> None:
    """Apply the collected grant statements."""
    for grant_statement in all_grants:
        # For global and specific columns:
        if create_roles:
            role = grant_statement.grantee
            if "UPDATE" in grant_statement.privileges or role.startswith("write_"):
                # Write users don't need .filtered
                _create_role_if_not_exists(conn, role, verbose=verbose, dry_run=dry_run)
            else:
                # Make sure both the regular and ".filtered" variant exists.
                # Users with direct connection to the database only inherit from the regular roles.
                # The DSO-API code can switch to the .filtered version,
                # which grants extra access to tables that have mandatoryFilterSets.
                if role.endswith(".filtered"):
                    app_role = role
                    role = role[: -len(".filtered")]
                else:
                    app_role = f"{role}.filtered"

                _create_role_if_not_exists(conn, role, verbose=verbose, dry_run=dry_run)
                _create_role_if_not_exists(
                    conn, app_role, inherits=role, verbose=verbose, dry_run=dry_run
                )

        _execute_grant(conn, grant_statement, verbose=verbose, dry_run=dry_run)


def _execute_grant(
    conn: Connection, grant_statement: _Grant, verbose: int = 1, dry_run: bool = False
) -> None:
    status_msg = "Skipped" if dry_run else "Executed"
    if verbose:
        logger.info("%s --> %s", status_msg, grant_statement)
    if not dry_run:
        sql_statement = f"""
            DO
            $$
                BEGIN
                    {grant_statement};
                EXCEPTION
                  WHEN undefined_table THEN
                    RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
                END
            $$
            """
        conn.execute(text(sql_statement))


def _create_role_if_not_exists(
    conn: Connection,
    role: str,
    inherits: str | None = None,
    verbose: int = 1,
    dry_run: bool = False,
) -> None:
    """Wrap the create role statement in an anonymous code block.

    Reason is to be able to catch exceptions.
    Don't break out of the session just because the role already exists
    """
    if role not in existing_roles:
        create_role_statement = f'CREATE ROLE "{role}"'
        if inherits:
            # PostgreSQL note, there are 2 syntax versions:
            # - "CREATE ROLE child IN ROLE parent" - this adds a member.
            # - "CREATE ROLE parent ROLE child"    - this declares a group, with initial members.
            create_role_statement = f'{create_role_statement} IN ROLE "{inherits}"'

        sql_statement = f"""
            DO $$
            BEGIN
            {create_role_statement};
            EXCEPTION
             WHEN duplicate_object
             THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
             WHEN undefined_object
             THEN RAISE NOTICE '%, skipping', SQLERRM USING ERRCODE = SQLSTATE;
            END
            $$
        """
        if verbose:
            status_msg = "Skipped" if dry_run else "Executed"
            logger.info("%s --> %s", status_msg, create_role_statement)
        if not dry_run:
            conn.execute(text(sql_statement))
        existing_roles.add(role)


def _get_sequence_name(conn: Connection, table: DatasetTableSchema) -> str | None:
    """Find the autoincrement sequence of a table."""
    if not table.is_autoincrement:
        return None

    column = table.identifier_fields[0].db_name  # always 1 field for autoincrement.
    key = (table.db_name, column)
    try:
        # Can't use lru_cache() as it caches 'session' too.
        return existing_sequences[key]
    except KeyError:
        row = conn.execute(
            text("SELECT pg_get_serial_sequence(:table, :column)"),
            {"table": table.db_name, "column": column},
        ).first()
        value = (
            row[0].replace("public.", "").replace('"', "")
            if row is not None and row[0] is not None
            else None
        )
        if not value:
            logger.debug("No sequence found for %s.%s", table.db_name, column)
        existing_sequences[key] = value
        return value


def _get_all_role_names(conn: Connection) -> list[str]:
    """Find all roles currently used in the database."""
    result = conn.execute(
        text(
            r"""
            SELECT rolname
            FROM pg_roles
            WHERE rolname LIKE 'scope\_%'
               OR rolname LIKE 'write\_%'
            """
        )
    )
    return [row[0] for row in result]


def _filter_grantees(
    scopes: frozenset[Scope] | frozenset[str], only_role: str | None, only_scope: str | None
) -> list[str]:
    """Determine which roles to assign for the scopes.
    This limits the granted scopes to the limitations given by CLI parameters.

    NOTE: The 'profiles' currently give a list of string scope names,
    while the 'dataset' objects already reference Scope objects.
    """
    if only_role is None:
        return [_scope_to_role(_scope) for _scope in scopes]
    elif any(s.id == only_scope if isinstance(s, Scope) else s == only_scope for s in scopes):
        return [only_role]
    else:
        return []


def _scope_to_role(scope: Scope | str) -> str:
    """Return rolename for the postgres database."""
    id = scope.id if isinstance(scope, Scope) else scope
    return f"scope_{id.lower().replace('/', '_')}"



# Generated by Django 5.2.1 on 2025-05-13 10:25
from __future__ import annotations

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0020_remove_remote_datasets"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="dataset",
            name="is_default_version",
        ),
        migrations.RemoveField(
            model_name="dataset",
            name="version",
        ),
        migrations.AddField(
            model_name="dataset",
            name="default_version",
            field=models.CharField(default="v1", verbose_name="Default version"),
        ),
        migrations.CreateModel(
            name="DatasetVersion",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("version", models.CharField(default="v1", max_length=3)),
                (
                    "lifecycle_status",
                    models.CharField(
                        choices=[("E", "experimental"), ("S", "stable")], default="E"
                    ),
                ),
                (
                    "dataset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="versions",
                        to="datasets.dataset",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="datasettable",
            name="dataset_versions",
            field=models.ManyToManyField(related_name="tables", to="datasets.datasetversion"),
        ),
        migrations.AlterUniqueTogether(
            name="datasettable",
            unique_together={("dataset", "name", "version")},
        ),
        migrations.AlterField(
            model_name="datasettable",
            name="db_table",
            field=models.CharField(max_length=100, unique=True),
        ),
    ]



# Generated by Django 5.2.5 on 2025-09-29 14:58
from __future__ import annotations

from django.db import migrations, models

import schematools.contrib.django.validators


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0022_inline_tables_for_all_versions_in_schema_data"),
    ]

    operations = [
        migrations.CreateModel(
            name="Publisher",
            fields=[
                ("id", models.CharField(max_length=100, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
                (
                    "schema_data",
                    models.TextField(
                        validators=[schematools.contrib.django.validators.validate_json],
                        verbose_name="Amsterdam Schema Contents",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Scope",
            fields=[
                ("id", models.CharField(max_length=100, primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=100)),
                (
                    "schema_data",
                    models.TextField(
                        validators=[schematools.contrib.django.validators.validate_json],
                        verbose_name="Amsterdam Schema Contents",
                    ),
                ),
            ],
        ),
    ]



"""Internal utils, not meant to be used outside schematools."""

from __future__ import annotations

import functools
import weakref


def cached_method(*lru_args, **lru_kwargs):
    """A simple lru-cache per object.
    This removed the need for methodtools.lru_cache(), which uses wirerope for purity.
    The usage of wirerope started showing up as 5% of the request time,
    hence it's significant to remove.
    """

    def decorator(func):
        @functools.wraps(func)
        def initial_wrapped_func(self, *args, **kwargs):
            # Not storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)

            # Assigns to the self reference (preserving the cache), and optimizes the next access.
            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)

        return initial_wrapped_func

    return decorator



# Generated by Django 5.1.4 on 2025-12-02 10:48
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0024_alter_profile_id"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="datasetversion",
            name="lifecycle_status",
        ),
        migrations.AddField(
            model_name="datasetversion",
            name="status",
            field=models.CharField(
                choices=[("D", "under_development"), ("S", "stable"), ("U", "superseded")],
                default="D",
            ),
        ),
    ]



from __future__ import annotations

import csv
from collections.abc import Iterable
from datetime import datetime
from typing import IO

from geoalchemy2 import functions as gfunc  # ST_AsEWKT
from sqlalchemy import Column, MetaData, func, select
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import ClauseElement

from schematools.exports import BaseExporter
from schematools.naming import toCamelCase
from schematools.types import DatasetFieldSchema, DatasetSchema, DatasetTableSchema

metadata = MetaData()


class CsvExporter(BaseExporter):  # noqa: D101
    extension = "csv"

    def geo_modifier(field: DatasetFieldSchema, column):
        if not field.is_geo:
            return column
        return gfunc.ST_AsEWKT(column).label(field.db_name)

    def id_modifier(field: DatasetFieldSchema, column):
        # We need an extra check for old relation definitions
        # that are defined as plain strings in the amsterdam schema
        # and not as objects.
        if field.table.is_temporal and (
            field.is_composite_key
            or field.related_table is not None
            and field.related_table.is_temporal
        ):
            return func.split_part(column, ".", 1).label(field.db_name)
        return column

    def datetime_modifier(field: DatasetFieldSchema, column):
        if field.type == "string" and field.format == "date-time":
            return func.to_char(column, 'YYYY-MM-DD"T"HH24:MI:SS').label(field.db_name)
        return column

    processors = (geo_modifier, id_modifier, datetime_modifier)

    def write_rows(
        self,
        file_handle: IO[str],
        table: DatasetTableSchema,
        columns: Iterable[Column],
        temporal_clause: ClauseElement | None,
        srid: str,
    ):
        field_names = [c.name for c in columns]
        writer = csv.DictWriter(file_handle, field_names, extrasaction="ignore")
        # Use capitalize() on headers, because csv export does the same
        writer.writerow({fn: toCamelCase(fn).capitalize() for fn in field_names})

        query = select(*columns)
        if temporal_clause is not None:
            query = query.where(temporal_clause)
        if self.size is not None:
            query = query.limit(self.size)

        # Use server-side cursor with small batches
        with self.connection.execution_options(stream_results=True, max_row_buffer=1000).execute(
            query
        ) as result:
            for partition in result.mappings().partitions(size=1000):
                writer.writerows(partition)


def export_csvs(
    connection: Connection,
    dataset_schema: DatasetSchema,
    output: str,
    table_ids: list[str],
    scopes: list[str],
    size: int,
    temporal_date: datetime | None = None,
):
    """Utility function to wrap the Exporter."""
    exporter = CsvExporter(
        connection, dataset_schema, output, table_ids, scopes, size, temporal_date
    )
    exporter.export_tables()
